# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "OpenHydroponics"
copyright = '2024, Micke Prag'
author = 'Micke Prag'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "jupyter_sphinx",
    "sphinx.ext.autodoc",
    "myst_parser",
    "sphinx.ext.graphviz",
    "sphinx_click",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

add_module_names = True
modindex_common_prefix = ["openhydroponics"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_context = {}
html_theme = "sphinx_material"
html_static_path = ['_static']
html_title = "Start"

html_theme_options = {
    "repo_url": "https://gitlab.com/openhydroponics/sw/openhydroponics",
    "repo_type": "gitlab",
    "repo_name": "OpenHydroponics GitLab",
    "html_minify": True,
    "css_minify": True,
    "nav_title": "OpenHydroponics",
    "globaltoc_depth": 3,
    "globaltoc_collapse": False,
    "nav_links": [
        {
            "href": "https://lectronz.com/stores/mickeprag",
            "internal": False,
            "title": "Store",
        },
    ],
}

html_sidebars = {
    "**": [
        "globaltoc.html",
        "localtoc.html",
        "searchbox.html",
    ]
}



"""Message pipe."""

import asyncio
import time
from typing import Optional


class MsgPipe:
    """
    This is a pipe that can have multiple listeners.
    The main difference between this and a queue is:

    - If no clients are actively listening the messages are discarded.
      So no queue will build up if no one is interested in the message.
      This also means no maximum queue length is required since the pipe will only hold one item at a time. The producer
      will be blocked until all listeners has received the message.
    - Multiple listeners are allowed, and all listeners will get all messages.
    """

    def __init__(self):
        self._lock = asyncio.Lock()
        self._condition = asyncio.Condition()
        self._msg = None

    async def send(self, msg):
        """
        Send a message to through the pipe. All active listeners will get the message.
        If no listeners are currently listening on the pipe the message will silently be dropped.
        """

        # Use a lock to make sure all messages are sent to handlers
        # Else the listeners may miss one is messages are fed rapidly
        async with self._lock:
            async with self._condition:
                self._msg = msg
                self._condition.notify_all()
            # Make sure all listener are awake before releasing the lock
            await asyncio.sleep(0)

    async def wait_for_one(self, timeout: float):
        """Wait for next incoming message"""
        async with self._condition:
            await asyncio.wait_for(self._condition.wait(), timeout)
            return self._msg

    async def wait_for_messages(self, timeout: Optional[float]):
        """Wait for multiple incoming messages"""
        start = time.time()
        async with self._condition:
            while True:
                if timeout:
                    time_left = timeout - (time.time() - start)
                else:
                    time_left = None
                try:
                    await asyncio.wait_for(self._condition.wait(), time_left)
                    yield self._msg
                except asyncio.exceptions.TimeoutError:
                    # Timeout occured, we are done listening for messages
                    return



from abc import ABC, abstractmethod
import asyncio
import functools
import time
from typing import Any, AsyncGenerator, TypeVar, Union
from uuid import UUID

from openhydroponics.base.endpoint import Endpoint
from openhydroponics.signal import signal

NodeType = TypeVar("Node")


class NodeManagerBase(ABC):

    def __init__(self):
        self._nodes = {}

    def add_node(self, node_id: Any, node: NodeType):
        """
        Add a new node to the manager. This method is only meant to be called by a subclass implementing the NodeManager
        interface.

        :param node_id: The unique identifier for the node. What this is depends on the implementation.
        :type node_id: Any
        :param node: The node instance to be added
        :type node: Node

        :raises: ValueError: If a node with the given ID already exists
        """
        if node_id in self._nodes:
            raise ValueError(f"Node with ID {node_id} already exists")
        node.on_endpoint_value_changed.connect(
            functools.partial(self.on_node_endpoint_value_changed, node)
        )
        self._nodes[node_id] = node
        self.on_node_added(node)

    def get_node(self, uuid: Union[str, UUID]) -> Union[NodeType, None]:
        """
        Retrieves a node by its UUID.

        :param: uuid (Union[str, UUID]): The UUID of the node to retrieve, either as a string or a UUID object.

        :returns: Union[NodeType, None]: The node with the specified UUID if found, None otherwise.
        """
        if isinstance(uuid, str):
            uuid = UUID(uuid)
        for node in self._nodes.values():
            if node.uuid == uuid:
                return node
        return None

    def get_node_by_id(self, node_id: Any) -> Union[NodeType, None]:
        """
        Retrieves a node by its internal ID. This is typically used for internal management
        and is not meant to be used by external code.
        This method is only meant to be called by a subclass implementing the NodeManager
        interface.

        :param node_id: The internal ID of the node to retrieve.
        :type node_id: Any

        :returns: Union[NodeType, None]: The node with the given ID if found, or None if not found.
        """
        return self._nodes.get(node_id, None)

    @abstractmethod
    async def init(self):
        """
        Initialize the node manager.

        This asynchronous method performs initialization tasks for the node manager.
        It should be called once after the node manager instance is created.

        :returns: None
        """
        pass

    @signal
    def on_node_added(self, node: NodeType):
        """
        Signal emitted when a node is added to the manager.

        :param node: The node that was added.
        """

    @signal
    def on_node_endpoint_value_changed(
        self, node: NodeType, endpoint: Endpoint, value: float, scale: int
    ):
        """
        Signal emitted when the value of a node's endpoint changes.

        :param node: The identifier of the node where the change occurred.
        :param endpoint: The specific endpoint of the node that changed.
        :param value: The new value of the endpoint.
        :param scale: The scale or unit associated with the value.

        """

    async def request_node(
        self, uuid: Union[str, UUID], timeout_s: float = 2.0
    ) -> Union[NodeType, None]:
        """
        Attempts to retrieve a node with the specified UUID, if the node is not found it will wait for the node to be
        available for a specified timeout period.
        This method is useful for ensuring that the node is available before proceeding with further operations.

        :param uuid: The UUID of the node to retrieve, either as a string or UUID object.
        :type uuid: str | UUID
        :param timeout_s: The maximum time in seconds to wait for the node to be available. Defaults to 2.0 seconds.
        :type timeout_s: float

        :returns: Node | None: The node with the specified UUID if found within the timeout period, None otherwise.
        """
        timeout = time.time() + timeout_s
        while time.time() < timeout:
            node = self.get_node(uuid)
            if node:
                return node
            await asyncio.sleep(0.1)
        return None

    def __iter__(self):
        return iter(self._nodes.values())

    def __aiter__(self):
        return aiter(self.nodes())

    async def nodes(self) -> AsyncGenerator[NodeType, None]:
        """
        Asynchronously yields all registered nodes.

        If no nodes are currently available, this method will wait for 1 second
        to allow for potential heartbeats from nodes to be received before continuing.

        :yields: Each node registered with the manager.
        """
        if not self._nodes:
            # No nodes found. Sleep to wait for heartbeats
            await asyncio.sleep(1)
        for node in self._nodes.values():
            yield node



import asyncio
import logging
import time
from typing import Any
from uuid import UUID
from pypostcard.types import List, u8
from pypostcard.serde import to_postcard

from openhydroponics.base import NodeManagerBase, NodeBase
from openhydroponics.signal import signal


from .phyinterface import CanPhyIface
import openhydroponics.net.msg as Msg
from openhydroponics.net.msg import ArbitrationId
from openhydroponics.net.endpoint import get_endpoint_class

_LOG = logging.getLogger(__name__)


class Node(NodeBase):
    def __init__(
        self, node_id: int, uuid: UUID, manager: NodeManagerBase, phy_iface: CanPhyIface
    ):
        super().__init__(uuid)
        self._manager = manager
        self._node_id = node_id
        self._phy_iface = phy_iface
        self._last_heartbeat = time.time()

    async def interview(self):
        """
        Interview the node to gather information about its endpoints.

        This asynchronous method queries the node for its number of endpoints (if not already known),
        then iterates through each endpoint to gather its information.

        This method is called by the NodeManager when the node is first discovered and should normally not be called
        directly by the user.
        The interview process involves sending requests to the node and waiting for responses.
        """
        if self.number_of_endpoints == -1:
            self.send_rtr(Msg.NodeInfo)
            resp = await self.wait_for(Msg.NodeInfo)
            if not resp:
                return
            self.number_of_endpoints = resp.number_of_endpoints
        for endpoint in range(self.number_of_endpoints):
            endpoint_info = await self.send_and_wait(
                Msg.EndpointInfoRequest(endpoint_id=u8(endpoint))
            )
            if not endpoint_info:
                continue
            EndpointClass = get_endpoint_class(
                endpoint_info.endpoint_class, endpoint_info.endpoint_sub_class
            )
            instance = EndpointClass(self, endpoint)
            await instance.interview()
            self.add_endpoint(endpoint, instance)
        self.on_interview_done()

    @property
    def interviewed(self) -> bool:
        """
        Check if all expected endpoints for this node have been interviewed/discovered.

        :returns: True if all expected endpoints have been discovered (number_of_endpoints equals
            the actual count of endpoints), False otherwise.
        """
        return self.number_of_endpoints == len(self.endpoints)

    @property
    def node_id(self) -> int:
        """
        Returns the node id of this node on the CANbus network.

        The node_id is an integer that uniquely identifies this node in the network.

        :returns: The unique identifier of this node.
        """
        return self._node_id

    @signal()
    def on_interview_done(self):
        """
        Signal emitted when the interview process is completed.
        This method is typically used to perform any necessary actions or
        updates after the interview process has finished successfully.
        """

    async def send_and_wait(self, request: Msg.Msg):
        """
        Send a request message and wait for the corresponding response.

        This method sends a request message to the node and waits for the matching response.
        It automatically determines the expected response type based on the request's message ID.

        :param request: The request message to send. Must be of MsgType.Request.
        :returns: The response message received.
        :raises AssertionError: If the request is not of MsgType.Request or if no corresponding
            response message class exists for the request's message ID.
        :raises asyncio.TimeoutError: If the response is not received within the specified timeout period.
        """
        assert request.MSG_TYPE == Msg.MsgType.Request
        response = Msg.Msg.get_msg_cls(request.MSG_ID, Msg.MsgType.Response)
        assert response
        self.send_msg(request)
        return await self.wait_for(response)

    def send_msg(self, msg: Msg.Msg):
        """
        Send a message to the node.
        This method encodes the message using the appropriate arbitration ID and sends it
        through the physical interface.

        :param msg: The message to send. Must be of type Msg.
        """
        arb = ArbitrationId(
            prio=False,
            dst=self._node_id,
            master=True,  # We are the master
            src=self._manager.node_id,
            multiframe=False,
            msg_type=msg.MSG_TYPE,
            msg_id=msg.MSG_ID,
        )
        data = Msg.Msg.encode(msg)
        self._phy_iface.send_message(arb.encode(), data)

    def send_rtr(self, msg: Any):
        """
        Send a remote transmission request (RTR) to the node.
        """
        arb = ArbitrationId(
            prio=False,
            dst=self._node_id,
            master=True,  # We are the master
            src=self._manager.node_id,
            multiframe=False,
            msg_type=Msg.MsgType.Request,
            msg_id=msg.MSG_ID,
        )
        self._phy_iface.send_message(arb.encode(), b"", is_remote=True)

    async def set_config(self, config_no: int, config: Any):
        """
        Set the configuration for the node.
        This method sends a configuration request to the node with the specified configuration number
        and configuration data.

        :param config_no: The configuration number to set.
        :param config: The configuration data to send.
        """
        cfg = to_postcard(config)
        cfg = cfg + bytes([0] * (32 - len(cfg)))
        msg = Msg.EndpointConfigRequest(
            endpoint_id=u8(self.node_id),
            config_no=u8(config_no),
            config=List(list(cfg)),
        )
        self.send_msg(msg)

    async def wait_for(self, msg: Msg.Msg):
        """
        Wait for a specific message type from the node.
        This method listens for incoming messages and returns the first message that matches the
        specified message type.

        :param msg: The message type to wait for. Must be of type Msg.
        :returns: The received message of the specified type.
        :raises asyncio.TimeoutError: If no matching message is received within the specified timeout period.
        """
        arb = ArbitrationId(
            prio=False,
            dst=self._manager.node_id,
            master=False,
            src=self._node_id,
            multiframe=False,
            msg_type=msg.MSG_TYPE,
            msg_id=msg.MSG_ID,
        )
        try:
            frame = await self._phy_iface.wait_for(arb.encode())
            return Msg.Msg.decode(arb, frame.data)
        except asyncio.TimeoutError:
            _LOG.error("Timeout waiting for frame. Arb: %s", arb)
        return None



import asyncio
import functools
from typing import Optional, Tuple
import click

from dbus_next.constants import BusType

from openhydroponics.base.endpoint import Endpoint, EndpointClass
from openhydroponics.dbus import NodeManager, Node


class KeyValueType(click.ParamType):
    name = "key=value"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Tuple[str, str]:
        """
        Parses a 'key=value' string into a (key, value) tuple.
        """
        try:
            # Split only on the first '='
            key, val = value.split("=", 1)
            key = key.strip()
            val = val.strip()
            if not key or not val:  # Ensure neither part is empty after stripping
                raise ValueError("Both key and value are required.")
            return key, val
        except ValueError:
            # Use self.fail for Click-specific error handling
            self.fail(
                f"'{value}' is not a valid key=value string.",
                param,
                ctx,
            )


def make_sync(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper


def get_node_manager(ctx) -> NodeManager:
    """
    Get the NodeManager instance from the context.
    """
    node_manager = ctx.obj.get("node_manager")
    if not node_manager:
        node_manager = NodeManager(bus_type=ctx.obj["bus_type"])
        ctx.obj["node_manager"] = node_manager
    return node_manager


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
    auto_envvar_prefix="HYPO",
)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--node",
    type=click.UUID,
    help="Node UUID. If not set, this is read from the environmental variable HYPO_NODE",
)
@click.option(
    "--bus",
    type=click.Choice(["system", "session"], case_sensitive=False),
    default="system",
    help="Select the dbus bus to connect to.",
)
@click.pass_context
def cli(ctx, node, bus):
    ctx.ensure_object(dict)
    ctx.obj["node"] = node
    if bus == "system":
        ctx.obj["bus_type"] = BusType.SYSTEM
    else:
        ctx.obj["bus_type"] = BusType.SESSION


@cli.group()
def config():
    """Get and set node configuration"""


@config.command(name="set")
@click.argument("endpoint", type=int)
@click.argument("config", type=KeyValueType(), nargs=-1, metavar="name=value")
@click.pass_context
@make_sync
async def config_set(ctx, endpoint, config):
    """Set node endpoint configuration"""
    nm = get_node_manager(ctx)
    await nm.init()

    config = dict(config)
    node: Node = await nm.request_node(ctx.obj["node"])
    if not node:
        click.echo("Could not find node")
        return
    endpoint: Endpoint = node.get_endpoint(endpoint)
    if not endpoint:
        click.echo("Could not find endpoint")
        return
    try:
        await endpoint.set_config(config)
    except Exception as e:
        click.echo(f"Error setting config: {e}")


@cli.command()
@click.pass_context
@make_sync
async def ls(ctx):
    """List nodes"""
    nm = get_node_manager(ctx)
    await nm.init()
    node: Node
    async for node in nm:
        click.echo(f"Node {node.uuid}:")
        for endpoint in node:
            if endpoint.ENDPOINT_CLASS == EndpointClass.Output:
                click.echo(
                    f" - EP{endpoint.endpoint_id} {endpoint.ENDPOINT_CLASS.name}: {endpoint.__class__.__name__}"
                )
                continue
            value, scale = node.get_endpoint_value(endpoint.endpoint_id)
            click.echo(
                f" - EP{endpoint.endpoint_id} {endpoint.ENDPOINT_CLASS.name}: {endpoint.__class__.__name__} {value:.2f}"
            )


@cli.group()
@click.pass_context
@make_sync
async def output(ctx):
    """Node endpoint commands"""


@output.command(name="set")
@click.argument("endpoint", type=int)
@click.argument("value", type=float)
@click.pass_context
@make_sync
async def output_set(ctx, endpoint: int, value: float):
    """Set node endpoint output"""
    nm = get_node_manager(ctx)
    await nm.init()
    node: Node = await nm.request_node(ctx.obj["node"])
    if not node:
        click.echo("Could not find node")
        return
    endpoint: Endpoint = node.get_endpoint(endpoint)
    if not endpoint:
        click.echo("Could not find endpoint")
        return
    if endpoint.ENDPOINT_CLASS != EndpointClass.Output:
        click.echo(f"Endpoint {endpoint} is not an output endpoint")
        return
    await endpoint.set(value)


if __name__ == "__main__":
    cli()



import logging

from serde import serde
from pypostcard.types import f32, u8

from openhydroponics.net.msg import ActuatorOutput, EndpointClass
from openhydroponics.base import endpoint as EndpointBase

_LOG = logging.getLogger(__name__)


@serde
class ECConfigCalibration:
    value: f32


@serde
class ECConfigGain:
    value: u8


class ECEndpoint(EndpointBase.ECEndpoint):

    async def set_config(self, config):
        if "high" in config and "low" in config:
            raise ValueError(
                "Cannot not set high and low at the same time, calibration will be wrong"
            )
        if "high" in config:
            await self.node.set_config(
                EndpointBase.ECConfigType.HIGH,
                ECConfigCalibration(value=f32(config["high"])),
            )
        if "low" in config:
            await self.node.set_config(
                EndpointBase.ECConfigType.LOW,
                ECConfigCalibration(value=f32(config["low"])),
            )
        if "gain" in config:
            await self.node.set_config(
                EndpointBase.ECConfigType.GAIN, ECConfigGain(value=u8(config["gain"]))
            )


class VariableOutputEndpoint(EndpointBase.VariableOutputEndpoint):
    async def set(self, value: float):
        self.node.send_msg(
            ActuatorOutput(endpoint_id=u8(self.endpoint_id), value=f32(value))
        )


def get_endpoint_input_class(
    endpoint_input_class: EndpointBase.EndpointInputClass,
) -> EndpointBase.InputEndpoint:
    if endpoint_input_class == EndpointBase.EndpointInputClass.Temperature:
        return EndpointBase.TemperatureEndpoint
    if endpoint_input_class == EndpointBase.EndpointInputClass.Humidity:
        return EndpointBase.HumidityEndpoint
    if endpoint_input_class == EndpointBase.EndpointInputClass.EC:
        return ECEndpoint
    if endpoint_input_class == EndpointBase.EndpointInputClass.PH:
        return EndpointBase.PHEndpoint
    return EndpointBase.InputEndpoint


def get_endpoint_output_class(
    endpoint_output_class: EndpointBase.EndpointOutputClass,
) -> EndpointBase.OutputEndpoint:
    if endpoint_output_class == EndpointBase.EndpointOutputClass.Variable:
        return VariableOutputEndpoint
    return EndpointBase.OutputEndpoint


def get_endpoint_class(
    endpoint_class: EndpointBase.EndpointClass, endpoint_sub_class
) -> EndpointBase.Endpoint:
    if endpoint_class == EndpointClass.Input:
        return get_endpoint_input_class(endpoint_sub_class)
    if endpoint_class == EndpointClass.Output:
        return get_endpoint_output_class(endpoint_sub_class)
    return EndpointBase.Endpoint



import logging
from typing import Any

from dbus_next import Variant, DBusError

from openhydroponics.base import endpoint as EndpointBase
from openhydroponics.service import endpoint as Service

_LOG = logging.getLogger(__name__)


def variant_from_val(value: Any):
    if isinstance(value, bool):
        return Variant("b", value)
    if isinstance(value, int):
        return Variant("i", value)
    if isinstance(value, float):
        return Variant("d", value)
    if isinstance(value, str):
        return Variant("s", value)
    raise ValueError(f"Unsupported type {type(value)}")


class Endpoint:
    async def set_config(self, config):
        try:
            return await self.dbus_interface.call_set_config(
                {key: variant_from_val(value) for key, value in config.items()}
            )
        except DBusError as e:
            raise Exception(e.text)

    async def init(self):
        pass


class InputEndpoint(Endpoint):
    async def init(self):
        await super().init()
        value, scale = await self.dbus_interface.get_value()
        self._value = value
        self._scale = scale
        self.properties_interface.on_properties_changed(self.on_properties_changed)

    def on_properties_changed(
        self,
        interface_name: str,
        changed_properties: dict[str, Variant],
        invalidated_properties: dict[str, Variant],
    ):
        for changed, variant in changed_properties.items():
            if changed == "Value":
                value, scale = variant.value
                self._value = value
                self._scale = scale
                self.on_value_changed(value, scale)


class VariableOutputEndpoint(EndpointBase.VariableOutputEndpoint):
    async def set(self, value: float):
        await self.dbus_interface.call_set(value)


MAPPING = {
    Service.InputEndpointInterface.DBUS_INTERFACE: EndpointBase.InputEndpoint,
    Service.TemperatureEndpointInterface.DBUS_INTERFACE: EndpointBase.TemperatureEndpoint,
    Service.HumidityEndpointInterface.DBUS_INTERFACE: EndpointBase.HumidityEndpoint,
    Service.ECEndpointInterface.DBUS_INTERFACE: EndpointBase.ECEndpoint,
    Service.PHEndpointInterface.DBUS_INTERFACE: EndpointBase.PHEndpoint,
    Service.OutputEndpointInterface.DBUS_INTERFACE: EndpointBase.OutputEndpoint,
    Service.VariableOutputEndpointInterface.DBUS_INTERFACE: VariableOutputEndpoint,
}


async def get_endpoint_class(proxy_object, node) -> EndpointBase.Endpoint:
    interface = proxy_object.get_interface(
        "com.openhydroponics.EndpointMetaDataInterface"
    )
    endpoint_id = await interface.get_endpoint_id()

    endpoint_interface_name = await interface.get_endpoint_interface()
    endpoint_interface = proxy_object.get_interface(endpoint_interface_name)
    properties_interface = proxy_object.get_interface("org.freedesktop.DBus.Properties")

    cls = MAPPING.get(endpoint_interface_name, EndpointBase.Endpoint)
    if cls.ENDPOINT_CLASS == EndpointBase.EndpointClass.Input:
        new_cls = type(cls.__name__, (InputEndpoint, cls), {})
    else:
        new_cls = type(cls.__name__, (Endpoint, cls), {})
    obj = new_cls(node, endpoint_id)
    obj.dbus_interface = endpoint_interface
    obj.properties_interface = properties_interface
    return obj



import asyncio
from functools import update_wrapper
import logging

_LOG = logging.getLogger(__name__)


class SignalInstance:
    def __init__(self):
        self._fn = None
        self._listeners = []

    def connect(self, callback):
        if asyncio.iscoroutinefunction(callback):
            self._listeners.append((callback, asyncio.get_event_loop()))
        else:
            self._listeners.append((callback, None))

    def __call__(self, *args, **kwargs):
        for listener, loop in self._listeners:
            try:
                if asyncio.iscoroutinefunction(listener):
                    asyncio.run_coroutine_threadsafe(listener(*args, **kwargs), loop)
                else:
                    listener(*args, **kwargs)
            except Exception as e:
                _LOG.error(f"Error in signal listener: {e}")


class SignalType:
    def __new__(cls, fn=None):
        instance = super().__new__(cls)
        if fn:
            return update_wrapper(instance, fn)
        return instance

    def __init__(self, fn=None):
        self._fn = fn

    def __get__(self, instance, _base):
        if not hasattr(instance, "_signals"):
            setattr(instance, "_signals", {})
        name = self._fn.__name__
        signal = instance._signals.get(name)
        if not signal:
            signal = SignalInstance()
            instance._signals[name] = signal
        return signal

    def __call__(self, fn):
        self._fn = fn
        return update_wrapper(self, fn)


signal = SignalType



import asyncio
import logging
from typing import Any, TypeVar
from uuid import UUID
import can
from plum import dispatch

from openhydroponics.base import NodeManagerBase
from openhydroponics.net.node import Node

from .phyinterface import CanPhyIface
import openhydroponics.net.msg as Msg
from openhydroponics.net.msg import ArbitrationId

_LOG = logging.getLogger(__name__)

NodeType = TypeVar("Node")


class NodeManager(NodeManagerBase):
    def __init__(self, phy_iface: CanPhyIface = None):
        super().__init__()
        self._loop = None
        self._node_id = 1
        self._last_node_id = 1

        self._phy_iface = phy_iface or CanPhyIface()
        self._phy_iface.add_listener(self._on_message_received)

    async def init(self):
        self._loop = asyncio.get_running_loop()
        self._phy_iface.set_event_loop(self._loop)

    @property
    def node_id(self) -> int:
        """
        Get the node id the manager is using on the CANbus.
        """
        return self._node_id

    def _on_message_received(self, msg: can.Message):
        arb = ArbitrationId.decode(msg.arbitration_id)
        msg = Msg.Msg.decode(arb, msg.data)
        if msg:
            self._handle_msg(arb, msg)
        else:
            _LOG.warning("Failed to decode message. Arb: %s", arb)

    @dispatch
    def _handle_msg(self, arb: ArbitrationId, msg: Msg.Heartbeat):
        node = self.get_node_by_id(arb.src)
        if node:
            # All good
            return
        node = Node(
            arb.src, UUID("00000000-00000000-00000000-00000000"), self, self._phy_iface
        )
        node.send_rtr(Msg.NodeInfo)

    @dispatch
    def _handle_msg(self, arb: ArbitrationId, msg: Msg.HeartbeatWithIdRequest):
        uuid = UUID(bytes=bytes(msg.uuid))
        if arb.src != 0:
            return
        node = self.get_node(uuid)
        if not node:
            node_id = self._last_node_id + 1
            _LOG.info(f"New node {uuid} found. Giving node id {node_id} to it")
            self._last_node_id = node_id
            node = Node(node_id, uuid, self, self._phy_iface)
            self.add_node(node_id, node)
        msg = Msg.IdSet(uuid=msg.uuid)
        node.send_msg(msg)

    @dispatch
    def _handle_msg(self, arb: ArbitrationId, msg: Msg.NodeInfo):
        uuid = UUID(bytes=bytes(msg.uuid))
        node = self.get_node_by_id(arb.src)
        if node:
            return
        _LOG.info(f"Node {uuid} detected.")
        node = Node(arb.src, uuid, self, self._phy_iface)
        node._number_of_endpoints = msg.number_of_endpoints
        self.add_node(arb.src, node)
        if self._loop:
            self._loop.create_task(node.interview())

    @dispatch
    def _handle_msg(self, arb: ArbitrationId, msg: Msg.SensorReading):
        node = self.get_node_by_id(arb.src)
        if not node:
            return
        endpoint = node.get_endpoint(msg.endpoint_id)
        if not endpoint:
            return
        endpoint.handle_sensor_reading(msg)

    @dispatch
    def _handle_msg(self, arb: ArbitrationId, msg: Any):
        pass



from typing import Union
from uuid import UUID

from dbus_next.aio import MessageBus, ProxyObject

from openhydroponics.base import NodeBase
from openhydroponics.base import endpoint as EndpointBase
from openhydroponics.dbus.endpoint import get_endpoint_class

BUS_NAME = "com.openhydroponics"


class Node(NodeBase):
    def __init__(self, uuid: UUID, proxy_object: ProxyObject):
        super().__init__(uuid)
        self._proxy_object = proxy_object
        self._interface = proxy_object.get_interface(
            "com.openhydroponics.NodeInterface"
        )

    def get_endpoint(self, endpoint_id: int) -> Union[EndpointBase.Endpoint, None]:
        endpoint = self._endpoints.get(endpoint_id)
        if not endpoint:
            return None
        return endpoint

    async def init(self, bus: MessageBus, introspection):
        for child in introspection.nodes:
            path = f"{introspection.name}/{child.name}"
            await self._add_endpoint_from_path(bus, path)
        # Resort endpoints
        self._endpoints = dict(sorted(self._endpoints.items()))

    async def _add_endpoint_from_path(self, bus: MessageBus, path: str):
        introspection = await bus.introspect(BUS_NAME, path)
        proxy_object = bus.get_proxy_object(BUS_NAME, path, introspection)

        endpoint = await get_endpoint_class(proxy_object, self)
        self.add_endpoint(endpoint.endpoint_id, endpoint)



from abc import ABC
import functools
from typing import Dict, Tuple, Union
from uuid import UUID

from openhydroponics.signal import signal
from openhydroponics.base.endpoint import Endpoint


class NodeBase(ABC):
    def __init__(self, uuid: UUID):
        self._endpoints = {}
        self._uuid = uuid
        self._number_of_endpoints = -1

    def add_endpoint(self, endpoint_no: int, instance: Endpoint):
        """
        Adds an endpoint to this node and sets up its value change tracking. This method is meant to be called
        by the node subclass when a new endpoint is added to the node.

        :param endpoint_no: A unique identifier for the endpoint within this node.
        :type endpoint_no: int
        :param instance: The endpoint instance to be added.
        :type instance: Endpoint
        """
        self._endpoints[endpoint_no] = instance
        instance.on_value_changed.connect(
            functools.partial(self.on_endpoint_value_changed, instance)
        )
        self.on_endpoint_added(instance)

    @property
    def endpoints(self) -> Dict[int, Endpoint]:
        """
        Returns the endpoints collection of this node.

        :returns: The collection of endpoint objects registered with this node.
        """
        return self._endpoints

    def get_endpoint(self, endpoint_id: int) -> Union[Endpoint, None]:
        """
        Retrieves an endpoint by its ID.

        :param endpoint_id: The ID of the endpoint to retrieve.
        :type endpoint_id: int

        :returns: The endpoint with the specified ID if it exists, None otherwise.
        """
        return self._endpoints.get(endpoint_id)

    def get_endpoint_value(self, endpoint_id: int) -> Tuple[float, int]:
        """
        Retrieves the value and scale of a specific endpoint by its ID.

        :param endpoint_id: The unique identifier of the endpoint to retrieve.

        :returns: A tuple containing the endpoint value (float or None if endpoint not found) and the endpoint scale
            (int or None if endpoint not found)
        """
        endpoint = self._endpoints.get(endpoint_id)
        if not endpoint:
            return (None, None)
        return endpoint.value, endpoint.scale

    @property
    def number_of_endpoints(self) -> int:
        """
        Returns the number of endpoints provided by this node.

        :returns: The number of endpoints that this node exposes.
        """
        return self._number_of_endpoints

    @number_of_endpoints.setter
    def number_of_endpoints(self, value: int):
        self._number_of_endpoints = value

    @signal
    def on_endpoint_added(self, endpoint: Endpoint):
        """
        Signal emitted when a new endpoint is added.

        :param endpoint: The endpoint object that has been added.
        :type endpoint: Endpoint
        """

    @signal
    def on_endpoint_value_changed(endpoint: Endpoint, value: float, scale: int):
        """
        Signal emitted when an endpoint's value changes.

        :param endpoint: The endpoint whose value has changed.
        :type endpoint: Endpoint
        :param value: The new value of the endpoint.
        :type value: float
        :param scale: The scaling factor or unit for the value.
        :type scale: int
        """
        pass

    @property
    def uuid(self) -> UUID:
        """
        Returns the UUID of the node.
        """
        return self._uuid

    def __iter__(self):
        return iter(self._endpoints.values())



from enum import IntEnum
import logging
from typing import Any

from openhydroponics.signal import signal

_LOG = logging.getLogger(__name__)


class EndpointClass(IntEnum):
    NotSupported = 0
    Input = 1
    Output = 2


class EndpointInputClass(IntEnum):
    NotSupported = 0
    Temperature = 1
    Humidity = 2
    EC = 3
    PH = 4


class EndpointOutputClass(IntEnum):
    NotSupported = 0
    Variable = 1
    Binary = 2


class Endpoint:
    ENDPOINT_CLASS = EndpointClass.NotSupported

    def __init__(self, node, endpoint_id: int):
        self._node = node
        self._endpoint_id: int = endpoint_id

    @property
    def endpoint_id(self) -> int:
        """
        Get the endpoint number for this endpoint.

        :returns: The endpoint's identifier that distinguishes it from other endpoints in this node.
        """
        return self._endpoint_id

    async def interview(self):
        """
        Conducts an interactive interview process for endpoint configuration.

        This asynchronous method prompts for and collects necessary information
        to set up or configure the endpoint.
        """
        pass

    @property
    def node(self):
        """
        Returns the node associated with this endpoint.
        """
        return self._node

    @signal()
    def on_value_changed(self, value: float, scale: int):
        """
        Signal emitted when a sensor reading is received.

        :param value: The value of the sensor reading.
        :param scale: The scale of the sensor reading.
        """

    async def set_config(self, config: dict[str, Any]):
        """
        Set configuration parameters for this endpoint.

        This method updates the endpoint's configuration with the provided dictionary of settings.

        :param config: A dictionary containing configuration parameters where keys are parameter names and values are the
            parameter values. This can include settings like thresholds, calibration values, or other endpoint-specific
            configurations.
        """
        pass


class InputEndpoint(Endpoint):
    ENDPOINT_CLASS = EndpointClass.Input
    INPUT_CLASS = EndpointInputClass.NotSupported

    def __init__(self, node, endpoint_id):
        super().__init__(node, endpoint_id)
        self._value = None
        self._scale = None

    def handle_sensor_reading(self, msg):
        self._value = msg.value
        self._scale = msg.scale
        self.on_value_changed(msg.value, msg.scale)

    @property
    def value(self):
        return self._value

    @property
    def scale(self):
        return self._scale


class TemperatureEndpoint(InputEndpoint):
    INPUT_CLASS = EndpointInputClass.Temperature


class HumidityEndpoint(InputEndpoint):
    INPUT_CLASS = EndpointInputClass.Humidity


class OutputEndpoint(Endpoint):
    ENDPOINT_CLASS = EndpointClass.Output
    OUTPUT_CLASS = EndpointOutputClass.NotSupported


class VariableOutputEndpoint(OutputEndpoint):
    OUTPUT_CLASS = EndpointOutputClass.Variable


class ECConfigType(IntEnum):
    LOW = 0
    HIGH = 1
    GAIN = 2


class ECEndpoint(InputEndpoint):
    INPUT_CLASS = EndpointInputClass.EC


class PHEndpoint(InputEndpoint):
    INPUT_CLASS = EndpointInputClass.PH



from uuid import UUID

from dbus_next.aio import MessageBus
from dbus_next.constants import BusType

from openhydroponics.base import NodeManagerBase
from openhydroponics.dbus import Node

BUS_NAME = "com.openhydroponics"


class NodeManager(NodeManagerBase):

    def __init__(self, bus_type: BusType = BusType.SYSTEM):
        super().__init__()
        self._bus = None
        self._bus_type: BusType = bus_type
        self._initialized = False

    async def init(self):
        if self._initialized:
            return
        self._bus = await MessageBus(bus_type=self._bus_type).connect()

        introspection = await self._bus.introspect(
            BUS_NAME, "/com/openhydroponics/nodes"
        )

        proxy_object = self._bus.get_proxy_object(
            BUS_NAME, "/com/openhydroponics/nodes", introspection
        )

        self._interface = proxy_object.get_interface("com.openhydroponics.NodeManager")
        self._interface.on_node_added(self._add_node_from_path)

        for child in introspection.nodes:
            path = f"{introspection.name}/{child.name}"
            await self._add_node_from_path(path)
        self._initialized = True

    async def _add_node_from_path(self, object_path: str):
        introspection = await self._bus.introspect(BUS_NAME, object_path)
        proxy_object = self._bus.get_proxy_object(BUS_NAME, object_path, introspection)

        interface = proxy_object.get_interface("com.openhydroponics.NodeInterface")
        uuid = UUID(await interface.get_uuid())
        node = Node(uuid, proxy_object)
        await node.init(self._bus, introspection)
        self.add_node(uuid, node)



from pypostcard.types import List, deserialize, i16, i8, serialize, u16, u8
import pytest

from openhydroponics.msg import ArbitrationId


@pytest.mark.parametrize(
    "raw, arb",
    [
        (
            0x01800000,
            ArbitrationId(
                prio=False,
                dst=12,
                master=False,
                src=0,
                multiframe=False,
                msg_type=0,
                msg_id=0,
            ),
        ),
        (
            0x05000000,
            ArbitrationId(
                prio=False,
                dst=40,
                master=False,
                src=0,
                multiframe=False,
                msg_type=0,
                msg_id=0,
            ),
        ),
        (
            0x06800000,
            ArbitrationId(
                prio=False,
                dst=52,
                master=False,
                src=0,
                multiframe=False,
                msg_type=0,
                msg_id=0,
            ),
        ),
        (
            0x00004801,
            ArbitrationId(
                prio=False,
                dst=0,
                master=False,
                src=2,
                multiframe=False,
                msg_type=2,
                msg_id=1,
            ),
        ),
    ],
)
def test_arbitration(raw: int, arb: ArbitrationId):
    decoded_arb = ArbitrationId.decode(raw)
    assert decoded_arb == arb
    assert arb.encode() == raw



from dbus_next.service import ServiceInterface, method, dbus_property
from dbus_next import DBusError, ErrorType, PropertyAccess

from openhydroponics.base.endpoint import (
    Endpoint,
    EndpointClass,
    EndpointInputClass,
    EndpointOutputClass,
    OutputEndpoint,
)


class EndpointInterface(ServiceInterface):
    DBUS_INTERFACE = "com.openhydroponics.EndpointInterface"

    def __init__(self, endpoint: Endpoint):
        super().__init__(self.DBUS_INTERFACE)
        self._endpoint = endpoint

    @method()
    async def SetConfig(self, config: "a{sv}"):
        try:
            # Unpack variants
            config = {key: variant.value for key, variant in config.items()}
            return await self._endpoint.set_config(config)
        except Exception as e:
            raise DBusError(ErrorType.SERVICE_ERROR, str(e))


class InputEndpointInterface(EndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.InputInterface"

    def __init__(self, endpoint: Endpoint):
        super().__init__(endpoint)
        self._endpoint.on_value_changed.connect(self._on_value_changed)

    @dbus_property(access=PropertyAccess.READ)
    def Value(self) -> "(dy)":
        if not self._endpoint.value:
            return [0.0, 0]
        return [self._endpoint.value, self._endpoint.scale]

    def _on_value_changed(self, value: float, scale: int):
        self.emit_properties_changed({"Value": [value, scale]})


class TemperatureEndpointInterface(InputEndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.TemperatureInterface"


class HumidityEndpointInterface(InputEndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.HumidityInterface"


class ECEndpointInterface(InputEndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.ECInterface"


class PHEndpointInterface(InputEndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.PHInterface"


class OutputEndpointInterface(EndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.OutputInterface"


class VariableOutputEndpointInterface(OutputEndpointInterface):
    DBUS_INTERFACE = "com.openhydroponics.endpoint.VariableOutputInterface"

    @method()
    async def Set(self, value: "d"):
        await self._endpoint.set(value)


def wrap_input_endpoint(endpoint: Endpoint) -> InputEndpointInterface:
    if endpoint.INPUT_CLASS == EndpointInputClass.Temperature:
        return TemperatureEndpointInterface(endpoint)
    if endpoint.INPUT_CLASS == EndpointInputClass.Humidity:
        return HumidityEndpointInterface(endpoint)
    if endpoint.INPUT_CLASS == EndpointInputClass.EC:
        return ECEndpointInterface(endpoint)
    if endpoint.INPUT_CLASS == EndpointInputClass.PH:
        return PHEndpointInterface(endpoint)
    return InputEndpointInterface(endpoint)


def wrap_output_endpoint(endpoint: OutputEndpoint) -> OutputEndpointInterface:
    if endpoint.OUTPUT_CLASS == EndpointOutputClass.Variable:
        return VariableOutputEndpointInterface(endpoint)
    return OutputEndpointInterface(endpoint)


def wrap_endpoint(endpoint: Endpoint) -> EndpointInterface:
    if endpoint.ENDPOINT_CLASS == EndpointClass.Input:
        return wrap_input_endpoint(endpoint)
    if endpoint.ENDPOINT_CLASS == EndpointClass.Output:
        return wrap_output_endpoint(endpoint)
    return EndpointInterface(endpoint)



import asyncio
import errno
import logging
import traceback
from typing import AsyncIterator, Callable, Awaitable, List, Optional
import can
import serial.tools.list_ports

from openhydroponics.msg_pipe import MsgPipe

_LOG = logging.getLogger(__name__)

# Suppress info and debug log messages from can module
logging.getLogger("can.interfaces.socketcan.socketcan").setLevel(logging.ERROR)

class Listener(can.Listener):
    def __init__(self, phy_iface):
        self._phy_iface = phy_iface

    def on_message_received(self, msg):
        self._phy_iface.on_message_received(msg)

    def on_error(self, exc: Exception) -> None:
        if not isinstance(exc, can.CanOperationError):
            _LOG.error(f"Unknown error: {exc}")
            return
        if exc.error_code == errno.ENETDOWN:
            _LOG.error("Network is down, reconnect bus...")
            self._phy_iface.reconnect()


class CanPhyIface:
    MsgCallback = Callable[[bytes], Awaitable[None]]

    def __init__(self):
        self._loop = None
        self._bus = None
        self._listeners: List[CanPhyIface.MsgCallback] = []
        self._msg_pipe = MsgPipe()
        self._init_kwargs = {
            "interface": "socketcan",
            "channel": "can0",
            "bitrate": 1000000,
            "receive_own_messages": False,
            "fd": True,
        }

        self._listener = Listener(self)
        self._bus = can.Bus(**self._init_kwargs)
        self._notifier = can.Notifier(self._bus, [self._listener])

    def add_listener(self, callback: MsgCallback) -> None:
        """Add a listener callback that will be called on for every message received"""
        self._listeners.append(callback)

    def reconnect(self):
        """Reconnect the bus"""
        async def do_it():
            if self._bus:
                self._notifier.stop()
                self._bus.shutdown()
                await asyncio.sleep(1)
            self._bus = can.Bus(**self._init_kwargs)
            self._notifier = can.Notifier(self._bus, [self._listener])

        if self._loop and self._loop.is_running():
            asyncio.run_coroutine_threadsafe(do_it(), self._loop)
        else:
            _LOG.warning("No loop running, cannot reconnect to bus")

    def on_message_received(self, msg: can.Message) -> None:
        # LOG.warn(f"Received message: {msg}")
        for listener in self._listeners:
            try:
                listener(msg)
            except Exception as exception:  # pylint: disable=broad-except
                _LOG.exception(exception)
                traceback.print_exception(exception)
        if self._loop and self._loop.is_running():
            asyncio.run_coroutine_threadsafe(self._msg_pipe.send(msg), self._loop)

    def send_message(self, arb: int, msg: bytes, is_remote=False) -> None:
        """Send a message over the CAN bus"""
        try:
            if is_remote:
                self._bus.send(
                    can.Message(arbitration_id=arb, data=b"", is_fd=True, is_remote_frame=False)
                )
            else:
                msg = CanPhyIface.pad_data(msg)
                self._bus.send(can.Message(arbitration_id=arb, data=msg, is_fd=True))
        except:
            LOG.error("Error sending message!!!!!!!!!!")


    def set_event_loop(self, loop):
        self._loop = loop

    async def wait_for(self, arb: int, timeout: Optional[float] = 2.0) -> can.Message:
        """Listen on incoming can frames and wait for a specific arbitration id"""
        async for msg in self.wait_for_many(arb, timeout):
            return msg

    async def wait_for_many(
        self, arb: int, timeout: Optional[float] = 2.0
    ) -> AsyncIterator[can.Message]:
        """Listen on incoming can frames and wait for multiple frames with a specific arbitration id"""

        async for msg in self._msg_pipe.wait_for_messages(timeout):
            if msg.arbitration_id == arb:
                yield msg
        raise asyncio.exceptions.TimeoutError()

    @staticmethod
    def get_slcan_channel():
        SUPPORTED_SLCAN_VIDS = [
            0xAD50,  # CANABLE
            0x16D0,  # CANABLE 2.0
            0x1E3,  # KORLAN_USB2CAN
        ]
        for port in serial.tools.list_ports.comports():
            if port.vid in SUPPORTED_SLCAN_VIDS:
                return port.device
        LOG.error("No supported slcan device found")
        return None

    @staticmethod
    def pad_data(data: bytes) -> bytes:
        """Pad data to valid CAN FD frame length"""
        if len(data) <= 8:
            return data
        for fdlen in (12, 16, 20, 24, 32, 48, 64):
            if fdlen >= len(data):
                return data.ljust(fdlen, b"\x00")
        raise ValueError("Data too long for CAN FD frame")



import functools
import logging
import asyncio

import click
from dbus_next.aio import MessageBus
from dbus_next.service import ServiceInterface, dbus_property
from dbus_next.constants import BusType, PropertyAccess
from dbus_next.service import signal

from openhydroponics.base.endpoint import Endpoint
from openhydroponics.net import Node, NodeManager
from openhydroponics.service.endpoint import wrap_endpoint


class NodeInterface(ServiceInterface):
    def __init__(self, bus: MessageBus, node: Node):
        super().__init__("com.openhydroponics.NodeInterface")
        self._bus = bus
        self._node = node
        self._endpoints = []
        for endpoint in node:
            endpoint_interface = EndpointInterface(self, endpoint)
            self._endpoints.append(endpoint_interface)
            self._bus.export(endpoint_interface.object_path, endpoint_interface)
            self._bus.export(
                endpoint_interface.object_path, endpoint_interface.wrapped_endpoint
            )

    @dbus_property(access=PropertyAccess.READ)
    def Interviewed(self) -> "b":
        return self._node.interviewed

    @property
    def object_path(self):
        return f"/com/openhydroponics/nodes/{str(self._node.uuid).replace('-', '_')}"

    @dbus_property(access=PropertyAccess.READ)
    def UUID(self) -> "s":
        return str(self._node.uuid)


class EndpointInterface(ServiceInterface):
    def __init__(self, node: NodeInterface, endpoint: Endpoint):
        super().__init__("com.openhydroponics.EndpointMetaDataInterface")
        self._node = node
        self._endpoint = endpoint
        self._wrapped_endoint = wrap_endpoint(endpoint)

    @property
    def wrapped_endpoint(self):
        return self._wrapped_endoint

    @dbus_property(access=PropertyAccess.READ)
    def EndpointClass(self) -> "u":
        return self._endpoint.ENDPOINT_CLASS

    @dbus_property(access=PropertyAccess.READ)
    def EndpointId(self) -> "u":
        return self._endpoint.endpoint_id

    @dbus_property(access=PropertyAccess.READ)
    def EndpointInterface(self) -> "s":
        return self._wrapped_endoint.DBUS_INTERFACE

    @property
    def object_path(self):
        return f"{self._node.object_path}/{self._endpoint.endpoint_id}"


class NodeManagerInterface(ServiceInterface):
    def __init__(self, bus):
        super().__init__("com.openhydroponics.NodeManager")
        self._bus: MessageBus = bus
        self._nm: NodeManager = NodeManager()
        self._nm.on_node_added.connect(self._on_node_added)
        self._nodes = []

    @signal()
    def NodeAdded(self, object_path: str) -> "o":
        return object_path

    async def init(self):
        await self._nm.init()

    def _on_node_added(self, node: Node):
        if not node.interviewed:
            # Re call ourselves when the interview is done
            node.on_interview_done.connect(
                functools.partial(self._on_node_added, node=node)
            )
            return
        node_interface = NodeInterface(self._bus, node)
        self._nodes.append(node_interface)
        self._bus.export(node_interface.object_path, node_interface)
        self.NodeAdded(node_interface.object_path)


async def main(bus_type: str):
    if bus_type == "system":
        bus = BusType.SYSTEM
    else:
        bus = BusType.SESSION

    logging.basicConfig(
        format="%(asctime)s %(name)-20s %(levelname)-7s: %(message)s",
        level=logging.DEBUG,
        datefmt="%H:%M:%S",
    )
    bus = await MessageBus(bus_type=bus).connect()

    interface = NodeManagerInterface(bus)
    await interface.init()

    bus.export("/com/openhydroponics/nodes", interface)
    await bus.request_name("com.openhydroponics")

    await bus.wait_for_disconnect()


@click.command()
@click.option(
    "--bus",
    type=click.Choice(["system", "session"], case_sensitive=False),
    default="system",
    help="Select the dbus bus to connect to.",
)
def daemon(bus):
    try:
        asyncio.run(main(bus))
    except KeyboardInterrupt:
        print("Exiting...")


if __name__ == "__main__":
    daemon()



from dataclasses import dataclass
from enum import IntEnum
import logging
from typing import Any

from serde import serde
from pypostcard.types import Enum, List, f32, u8, u16, u32
from pypostcard.serde import from_postcard, take_from_postcard, to_postcard

from openhydroponics.base.endpoint import EndpointClass

_LOG = logging.getLogger(__name__)
@dataclass
class ArbitrationId:
    prio: bool
    dst: int
    master: bool
    src: int
    multiframe: bool
    msg_type: int
    msg_id: int

    def encode(self):
        id = (
            (self.prio << 28)
            | (self.dst << 21)
            | (self.master << 20)
            | (self.src << 13)
            | (self.multiframe << 12)
            | (self.msg_type << 10)
            | self.msg_id
        )
        return id

    @staticmethod
    def decode(id: int):
        return ArbitrationId(
            prio=bool((id >> 28) & 0x1),
            dst=(id >> 21) & 0b1111111,
            master=bool((id >> 20) & 1),
            src=(id >> 13) & 0b1111111,
            multiframe=bool((id >> 12) & 1),
            msg_type=(id >> 10) & 0b11,
            msg_id=id & 0b1111111111,
        )


class MsgType(IntEnum):
    Request = 0
    Response = 1
    Value = 2


class MsgId(IntEnum):
    Heartbeat = 1
    HeartbeatWithIdRequest = 2
    IdSet = 3
    NodeInfo = 4
    EndpointInfo = 5
    SensorReading = 6
    ActuatorOutput = 7
    EndpointConfig = 8


class Msg:
    messages = {}

    def __init__(self, msg_id: MsgId, msg_type: MsgType) -> None:
        self._msg_id = msg_id
        self._msg_type = msg_type

    def __call__(self, cls):
        cls.MSG_ID = self._msg_id
        cls.MSG_TYPE = self._msg_type
        Msg.messages[(self._msg_id, self._msg_type)] = cls
        return serde(cls)

    @staticmethod
    def encode(msg: Any) -> bytes:
        return to_postcard(msg)

    @staticmethod
    def decode(arb: ArbitrationId, data: bytes):
        msg_id = arb.msg_id
        msg_type = arb.msg_type
        msg_cls = Msg.messages.get((msg_id, msg_type), None)
        if not msg_cls:
            _LOG.debug(f"Unknown message id: {msg_id} {msg_type}")
            return None
        return from_postcard(msg_cls, data)

    @staticmethod
    def get_msg_cls(msg_id: MsgId, msg_type: MsgType):
        return Msg.messages.get((msg_id, msg_type), None)


@Msg(MsgId.Heartbeat, MsgType.Value)
class Heartbeat:
    counter: u32


@Msg(MsgId.HeartbeatWithIdRequest, MsgType.Request)
class HeartbeatWithIdRequest:
    counter: u32
    uuid: List[u8, 16]


@Msg(MsgId.IdSet, MsgType.Request)
class IdSet:
    uuid: List[u8, 16]


@Msg(MsgId.NodeInfo, MsgType.Value)
class NodeInfo:
    uuid: List[u8, 16]
    number_of_endpoints: u8


@Msg(MsgId.EndpointInfo, MsgType.Request)
class EndpointInfoRequest:
    endpoint_id: u8


@Msg(MsgId.EndpointInfo, MsgType.Response)
class EndpointInfoResponse:
    id: u8
    endpoint_class: Enum[u8, EndpointClass]
    endpoint_sub_class: u8


@Msg(MsgId.SensorReading, MsgType.Value)
class SensorReading:
    endpoint_id: u8
    value: f32
    scale: u8


@Msg(MsgId.ActuatorOutput, MsgType.Request)
class ActuatorOutput:
    endpoint_id: u8
    value: f32


@Msg(MsgId.EndpointConfig, MsgType.Request)
class EndpointConfigRequest:
    endpoint_id: u8
    config_no: u8
    config: List[u8, 32]
